from abc import ABCMeta, abstractmethod
from typing import List

from ..config.data_type import DataType


class Warehouse(metaclass=ABCMeta):
    """Abstract Warehouse class"""

    conf_dict: dict

    def __init__(self, conf_dict):
        """Should Create connection"""
        self.conf_dict = conf_dict
        self.connect()

    @abstractmethod
    def connect(self):
        return

    @abstractmethod
    def create_schema(self, schema: str):
        """ Create schema or namespace if does not exist"""
        return

    @abstractmethod
    def create_table(self, schema: str, table: str, col_types: dict, non_null_columns: List[str]):
        """ Create table if does not exist"""
        return

    @abstractmethod
    def create_users_table(self, schema: str, col_types: dict, non_null_columns: List[str]):
        """ Create users table if does not exist"""
        return

    @abstractmethod
    def describe_table(self, chema: str, table: str):
        return

    @abstractmethod
    def add_column(self, schema: str, table: str, column: str, column_type: DataType, non_null_columns: List[str]):
        return

    @abstractmethod
    def insert_df(self, schema: str, table: str, df):
        return

    @abstractmethod
    def close(self):
        return



def flatten_json(y):
    out = {}

    def flatten(x, name=""):
        if type(x) is dict:
            for a in x:
                flatten(
                    x[a],
                    clean_event_key(name) + clean_event_key(a) + "_",
                )
        elif type(x) is list:
            i = 0
            for a in x:
                flatten(a, clean_event_key(name) + str(i) + "_")
                i += 1
        else:
            out[name[:-1]] = x

    flatten(y)
    return out


def clean_event_key(key):
    return key.strip().replace(" ", "").replace(":", "_").replace("-", "_")



import logging

import numpy as np
import pandas as pd

from ..config import event_fields, data_type


def get_datatypes(df):
    column_names = list(df.columns.values)
    column_datatypes = {}

    for c in column_names:
        first_good_value = first_valid_value(df, c)
        if first_good_value is None:
            # skip this column
            continue

        if df[c].dtype == object and isinstance(first_good_value, str):
            if c in event_fields.TIMESTAMP_FIELDS:
                column_datatypes[c] = data_type.DataType.DATETIME
            else:
                column_datatypes[c] = data_type.DataType.STRING
                df[c] = df[c].astype(str)
        elif df[c].dtype == np.float64 or float == type(first_good_value):
            column_datatypes[c] = data_type.DataType.FLOAT64
        elif df[c].dtype == np.int64 or int == type(first_good_value):
            column_datatypes[c] = data_type.DataType.INT64
        elif df[c].dtype == np.bool_ or bool == type(first_good_value):
            column_datatypes[c] = data_type.DataType.BOOLEAN
        elif "datetime64" in str(df[c].dtype):
            column_datatypes[c] = data_type.DataType.DATETIME
        else:
            raise Exception(
                f"Unknown type. dtype= {df[c].dtype}. column = {c}, first value = {first_good_value}, Python Type = {type(first_good_value)} "
            )

    return column_datatypes


def first_valid_value(df, column):
    """Returns None if no valid value exists else returns value"""
    valid_index = df[column].first_valid_index()
    if valid_index is None:
        return None
    logging.debug(
        f"valid_index = {valid_index}, column = {column}, df[column][{valid_index}] = {df[column][valid_index]}"
    )
    return df[column][valid_index]


def row_count(df):
    """Faster way to get length of df"""
    return len(df.index)


def empty(df):
    return row_count(df) == 0


def mark_nan_to_none(df, col_types):
    return df.where(pd.notnull(df), None)


def mark_string_na_to_default(df, col_types):
    for column_name, column_type in col_types.items():
        if column_type == data_type.DataType.STRING:
            df[column_name].fillna("_default", inplace=True)


def mark_int_na_to_default(df, col_types):
    for column_name, column_type in col_types.items():
        if column_type == data_type.DataType.INT64:
            df[column_name].fillna(0, inplace=True)


def mark_float_na_to_default(df, col_types):
    for column_name, column_type in col_types.items():
        if column_type in (data_type.DataType.FLOAT64, data_type.DataType.FLOAT32):
            df[column_name].fillna(0.0, inplace=True)


def cast_boolean_to_int(df, col_types):
    for column_name, column_type in col_types.items():
        if column_type == data_type.DataType.BOOLEAN:
            df[column_name].fillna(False, inplace=True)
            df[column_name] = df[column_name].astype(int)


def add_missing_columns(df, col_types):
    existing_cols = get_datatypes(df)
    for column_name, column_type in col_types.items():
        if column_name not in existing_cols:
            df[column_name] = None


def fix_data_types(df, df_dicts, expected_col_types):
    """ Fixes data types and returns misfits if not able to fix """
    df_col_types = get_datatypes(df)
    misfits = []

    for column_name, column_type in expected_col_types.items():
        if column_name not in df_col_types:
            continue

        if df_col_types[column_name] == expected_col_types[column_name]:
            continue
        else:
            if expected_col_types[column_name] == data_type.DataType.STRING:
                cast_to_str(column_name, df_dicts)
            elif expected_col_types[column_name] in data_type.INT_DATATYPES:
                if df_col_types[column_name] in data_type.INT_DATATYPES:
                    # Let us hope similar integers will be handled wisely by downstream
                    continue
                elif df_col_types[column_name] in data_type.FLOAT_DATATYPES:
                    cast_to_int(column_name, df_dicts)
                elif df_col_types[column_name] == data_type.DataType.STRING:
                    cast_to_int(column_name, df_dicts)
                else:
                    raise Exception(
                        f"Dont know how to handle. Column = {column_name}, Expected {expected_col_types[column_name]}, Actual {df_col_types[column_name]}"
                    )
            elif expected_col_types[column_name] in data_type.FLOAT_DATATYPES:
                if df_col_types[column_name] in data_type.FLOAT_DATATYPES:
                    # Let us hope similar float variations will be handled wisely by downstream
                    continue
                elif df_col_types[column_name] in data_type.INT_DATATYPES:
                    cast_to_float(column_name, df_dicts)
                else:
                    raise Exception(
                        f"Dont know how to handle. Column = {column_name}, Expected {expected_col_types[column_name]}, Actual {df_col_types[column_name]}"
                    )
            else:
                raise Exception(
                    f"Dont know how to handle. Column = {column_name}, Expected {expected_col_types[column_name]}, Actual {df_col_types[column_name]}"
                )


def cast_to_float(column_name, df_dicts):
    for d in df_dicts:
        if d[column_name] is not None:
            d[column_name] = float(d[column_name])


def cast_to_str(column_name, df_dicts):
    for d in df_dicts:
        if d[column_name] is not None:
            d[column_name] = str(d[column_name])


def cast_to_int(column_name, df_dicts):
    for d in df_dicts:
        if d[column_name] is not None:
            d[column_name] = int(d[column_name])



import logging
import subprocess
import uuid

TMP_DIR_PREFIX = "seghouse"


def s3_copy(s3_path, local_dir_path):
    command = ["aws", "s3", "cp", s3_path, local_dir_path, "--recursive", "--exclude", "*", "--include", "*.gz"]
    logging.info(f"command = {command}")
    process = subprocess.run(command)
    process.check_returncode()


def download_gz_files(s3_dir):
    local_dir_name = str(uuid.uuid4()).replace("-", "")
    local_dir_path = f"/tmp/{TMP_DIR_PREFIX}-{local_dir_name}"
    subprocess.run(["mkdir", "-p", local_dir_path])

    logging.info(f"Copying files to {local_dir_path}")
    s3_copy(s3_dir, local_dir_path)

    return local_dir_path



import logging
import shutil

import click

from .config import configuration
from .jobs import send_to_warehouse
from .util import aws_wrapper

logging.basicConfig(level=logging.INFO)


@click.group()
def app():
    """Send SegmentSpec event files to Warehouses"""


@app.command()
@click.option("--config-file", "-cf", type=click.Path(exists=True))
@click.option("--s3-dir", "-s3d",
              help="S3 Directory. We will look for *.gz files in this directory. "
                   "Ensure that you have configured aws "
                   "credentials using aws cli. "
                   "Make sure that this directory contains files less than 100.")
@click.option("--source-dir", "-sd", type=click.Path(exists=True))
@click.option("--namespace", "-ns", required=True, help="Will be used to create database/namespace in warehouse", )
def send(config_file: str, s3_dir: str, source_dir: str, namespace: str):
    """Send Segment Files to different warehouses """
    logging.info(f"config_file={config_file}")
    app_conf = configuration.from_yaml(config_file)

    try:
        if s3_dir:
            source_dir = aws_wrapper.download_gz_files(s3_dir)

        job = send_to_warehouse.SendToWarehouseJob(app_conf, source_dir, namespace)
        job.execute()
    finally:
        if s3_dir:
            logging.info(f"Removing directory {source_dir}")
            shutil.rmtree(source_dir)



from .clickhouse import ClickHouse


def get_warehouse(warehouse_conf: dict):
    if "clickhouse" == warehouse_conf["type"]:
        return ClickHouse(warehouse_conf)
    else:
        raise WarehouseError(f"Unable to get warehouse of type {type}")


class WarehouseError(Exception):
    def __init__(self, message):
        self.message = message



import logging
from typing import Set, List

from clickhouse_driver import Client

from .warehouse import Warehouse
from ..config.data_type import DataType
from ..util import dataframe_util

SAMPLE_QUERY = "SELECT 1"
DT_TO_CH_DT = {
    DataType.UINT8: "UInt8",
    DataType.UINT16: "UInt16",
    DataType.UINT32: "UInt32",
    DataType.UINT64: "UInt64",
    DataType.UINT256: "UInt256",
    DataType.INT8: "Int8",
    DataType.INT16: "Int16",
    DataType.INT32: "Int32",
    DataType.INT64: "Int64",
    DataType.INT128: "Int128",
    DataType.INT256: "Int256",
    DataType.FLOAT32: "Float32",
    DataType.FLOAT64: "Float64",
    DataType.BOOLEAN: "UInt8",
    DataType.STRING: "String",
    DataType.DATE: "Date",
    DataType.DATETIME: "DateTime",
}


class ClickHouse(Warehouse):
    clickhouse_client: Client
    clickhouse_cluster: str
    created_tables: Set[str]

    def connect(self):
        self.clickhouse_client = Client(
            host=self.conf_dict["host"],
            port=self.conf_dict.get("port", 9000),
            user=self.conf_dict["user"],
            password=self.conf_dict["password"],
        )
        self.clickhouse_cluster = self.conf_dict.get("cluster")
        logging.info("connecting to ClickHouse")
        logging.info(f"Running sample query {SAMPLE_QUERY}")

        result = self.clickhouse_client.execute(SAMPLE_QUERY)
        logging.info(f"Result = {result}")

        self.created_tables = set()
        return True

    # @abstractmethod
    def create_schema(self, schema: str):
        """ Create schema or namespace if does not exist"""
        create_db_sql = f"CREATE DATABASE IF NOT EXISTS {schema}"
        if self.clickhouse_cluster:
            create_db_sql = f"{create_db_sql} ON CLUSTER {self.clickhouse_cluster}"

        result = self.clickhouse_client.execute(create_db_sql)
        logging.debug(f"Creating Database {schema}, result = {result}")

    # @abstractmethod
    def create_table(self, schema: str, table: str, col_types: dict, non_null_columns: List[str]):
        """ Create table if does not exist"""
        if f"{schema}.{table}" in self.created_tables:
            return

        if self.clickhouse_cluster:
            raise Exception("ClickHouse cluster is not yet implemented")
        else:
            column_type_defs = []
            for col_name, col_type in col_types.items():
                column_type_defs.append(self.to_ch_column_def(col_name, col_type, non_null_columns))

            sql = f"""
            CREATE TABLE IF NOT EXISTS {schema}.{table}
            (
                {', '.join(column_type_defs)}
            ) ENGINE = ReplacingMergeTree()
            PARTITION BY toDate(timestamp)
            ORDER BY (timestamp, message_id)
            """
        logging.debug(f"Running SQL = {sql}")
        result = self.clickhouse_client.execute(sql)
        logging.debug(f"Creating Table {schema}.{table}, result = {result}")

        self.created_tables.add(f"{schema}.{table}")

    def create_users_table(self, schema: str, col_types: dict, non_null_columns: List[str]):
        """ Create table if does not exist"""
        table = "users"
        if f"{schema}.{table}" in self.created_tables:
            return

        if self.clickhouse_cluster:
            raise Exception("ClickHouse cluster is not yet implemented")
        else:
            column_type_defs = []
            for col_name, col_type in col_types.items():
                column_type_defs.append(
                    self.to_ch_column_def(
                        col_name, col_type, non_null_columns
                    )
                )

            sql = f"""
            CREATE TABLE IF NOT EXISTS {schema}.{table}
            (
                {', '.join(column_type_defs)}
            ) ENGINE = ReplacingMergeTree(ver)
            PARTITION BY (toDate(timestamp), user_id)
            ORDER BY (user_id)
            """
        logging.debug(f"Running SQL = {sql}")
        result = self.clickhouse_client.execute(sql)
        logging.debug(f"Creating Table {schema}.{table}, result = {result}")

        self.created_tables.add(f"{schema}.{table}")

    @staticmethod
    def to_ch_column_def(
            column_name, column_type, non_null_columns=["received_at", "timestamp", "message_id"]
    ):
        ch_type = DT_TO_CH_DT.get(column_type)
        if ch_type is None:
            raise Exception(f"Unable to find ch_type for DT = {column_type}")
        if column_name not in non_null_columns:
            ch_type = f"Nullable({ch_type})"
        return f"{column_name} {ch_type}"

    # @abstractmethod
    def describe_table(self, schema: str, table: str):
        sql = f"DESCRIBE TABLE {schema}.{table}"
        logging.debug(f"Running SQL = {sql}")
        result = self.clickhouse_client.execute(sql)
        col_types = {}
        for x in result:
            col_types[x[0]] = self.ch_type_to_seghouse_type(x[1])
        return col_types

    @staticmethod
    def ch_type_to_seghouse_type(ch_type):
        if "UInt8" in ch_type:
            return DataType.UINT8
        elif "UInt16" in ch_type:
            return DataType.UINT16
        elif "UInt32" in ch_type:
            return DataType.UINT32
        elif "UInt64" in ch_type:
            return DataType.UINT64
        elif "UInt256" in ch_type:
            return DataType.UINT256
        elif "Int8" in ch_type:
            return DataType.INT8
        elif "Int16" in ch_type:
            return DataType.INT16
        elif "Int32" in ch_type:
            return DataType.INT32
        elif "Int64" in ch_type:
            return DataType.INT64
        elif "Int128" in ch_type:
            return DataType.INT128
        elif "Int256" in ch_type:
            return DataType.INT256
        elif "Float32" in ch_type:
            return DataType.FLOAT32
        elif "Float64" in ch_type:
            return DataType.FLOAT64
        elif "UInt8" in ch_type:
            return DataType.BOOLEAN
        elif "String" in ch_type:
            return DataType.STRING
        elif "DateTime" in ch_type:
            return DataType.DATETIME
        elif "Date" in ch_type:
            return DataType.DATE
        else:
            raise Exception(f"unable to convert ch_type {ch_type}")

    def add_column(self, schema: str, table: str, column: str, column_type: DataType, non_null_columns: List[str]):
        sql = f"ALTER TABLE {schema}.{table} ADD COLUMN IF NOT EXISTS {self.to_ch_column_def(column, column_type, non_null_columns)}"
        logging.debug(f"Running SQL = {sql}")
        result = self.clickhouse_client.execute(sql)
        logging.debug(
            f"Adding column to {schema}.{table}, {column}, {column_type} result = {result}"
        )

    def insert_df(self, schema: str, table: str, dataframe):
        df = dataframe.copy()
        col_types = dataframe_util.get_datatypes(df)
        dataframe_util.cast_boolean_to_int(df, col_types)
        # dataframe_util.mark_int_na_to_default(df, col_types)
        # dataframe_util.mark_float_na_to_default(df, col_types)

        table_column_types = self.describe_table(schema, table)
        dataframe_util.add_missing_columns(df, table_column_types)
        logging.debug(f"{table} table_column_types = {table_column_types}")

        df_dicts = df.to_dict("records")
        dataframe_util.fix_data_types(df, df_dicts, table_column_types)

        result = self.clickhouse_client.execute(
            f"INSERT INTO {schema}.{table} VALUES",
            df_dicts,
            types_check=True,
        )
        logging.info(f"Inserting DataFrame in {schema}.{table}, result = {result}")

    def close(self):
        return



from enum import Enum, unique


@unique
class DataType(Enum):
    "Lists all Data Types"
    UINT8 = "uint8"
    UINT16 = "uint16"
    UINT32 = "uint32"
    UINT64 = "uint64"
    UINT256 = "uint256"
    INT8 = "int8"
    INT16 = "int16"
    INT32 = "int32"
    INT64 = "int64"
    INT128 = "int128"
    INT256 = "int256"
    FLOAT32 = "float"
    FLOAT64 = "double"
    BOOLEAN = "boolean"
    STRING = "string"
    UUID = "uuid"
    DATE = "date"
    DATETIME = "datetime"
    ARRAY = "array"


INT_DATATYPES = (
    DataType.UINT8,
    DataType.UINT16,
    DataType.UINT32,
    DataType.UINT64,
    DataType.UINT256,
    DataType.INT8,
    DataType.INT16,
    DataType.INT32,
    DataType.INT64,
    DataType.INT128,
    DataType.INT256,
)

FLOAT_DATATYPES = (DataType.FLOAT64, DataType.FLOAT32)



from dataclasses import dataclass
from typing import List

import humps
import yaml


@dataclass(frozen=True, eq=True)
class App:
    """Class for recording app names with write keys."""

    write_key: str
    name: str

    def schema(self):
        return humps.decamelize(self.name)


yaml.add_path_resolver("!app", ["App"], dict)


@dataclass(frozen=True, eq=True)
class AppConf:
    """Top level configuration class"""

    apps: List[App]
    warehouses: List[dict]
    skip_fields: List[str]
    extra_timestamps: dict


def from_yaml(file_path: str):
    apps = set()
    skip_fields = []
    with open(file_path) as file:
        resolved_conf = yaml.load(file, Loader=yaml.FullLoader)
        for app_dict in resolved_conf.get("apps", []):
            apps.add(App(app_dict["write_key"], app_dict["name"]))
        for f in resolved_conf.get("skip_fields", []):
            skip_fields.append(f)
        extra_timestamps = resolved_conf.get("extra_timestamps", {})
    return AppConf(
        apps=list(apps), warehouses=resolved_conf["warehouses"], skip_fields=skip_fields,
        extra_timestamps=extra_timestamps
    )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['seghouse',
 'seghouse.config',
 'seghouse.jobs',
 'seghouse.util',
 'seghouse.warehouse']

package_data = \
{'': ['*']}

install_requires = \
['PyYAML>=5.3.1,<6.0.0',
 'click>=7.1.2,<8.0.0',
 'clickhouse-driver[lz4,numpy,zstd]>=0.2.0,<0.3.0',
 'pandas>=1.2.0,<2.0.0',
 'pyhumps>=1.6.1,<2.0.0',
 'tabulate>=0.8.7,<0.9.0']

entry_points = \
{'console_scripts': ['seghouse = seghouse.app:app']}

setup_kwargs = {
    'name': 'seghouse',
    'version': '0.8.0',
    'description': '',
    'long_description': None,
    'author': 'Dinesh Sawant',
    'author_email': 'dineshsawant300@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.7.1,<4.0.0',
}


setup(**setup_kwargs)



from .data_type import DataType

BASE_STRUCTURE = {
    "message_id": DataType.STRING,
    "anonymous_id": DataType.STRING,
    "received_at": DataType.DATETIME,
    "timestamp": DataType.DATETIME,
    "ip": DataType.STRING,
}

EVENT_SPECIFIC = {
    "original_event": DataType.STRING,
    "event": DataType.STRING,
}

TRACKS = dict(BASE_STRUCTURE)
TRACKS.update(EVENT_SPECIFIC)
TRACKS_ALLOWED_FIELD_PREFIXES = (
    "context_",
    "traits_",
    "geoip_",
)

IDENTITIES = dict(BASE_STRUCTURE)

PAGES = dict(BASE_STRUCTURE)

SCREENS = dict(BASE_STRUCTURE)

USERS = dict(BASE_STRUCTURE)
USER_SPECIFIC = {"user_id": DataType.STRING, "ver": DataType.INT64}
USERS.update(USER_SPECIFIC)

GROUPS = dict(BASE_STRUCTURE)

ALIASES = dict(BASE_STRUCTURE)

TRACKS_TABLE = "tracks"
SCREENS_TABLE = "screens"
IDENTITIES_TABLE = "identities"
PAGES_TABLE = "pages"
USERS_TABLE = "users"
ALIASES_TABLE = "aliases"
GROUPS_TABLE = "groups"
MISFITS_TABLE = "misfits"

DEFAULT_TABLES = [
    TRACKS_TABLE,
    SCREENS_TABLE,
    IDENTITIES_TABLE,
    PAGES_TABLE,
    USERS_TABLE,
    ALIASES_TABLE,
    GROUPS_TABLE,
    MISFITS_TABLE
]



import gzip
import json
import logging
from dataclasses import dataclass
from os import listdir
from os.path import isfile, join
from typing import List

import humps
import pandas as pd

from ..config import default_table_structure
from ..config import event_fields
from ..config.configuration import AppConf
from ..util import json_util, dataframe_util
from ..warehouse import factory as whf, warehouse as wh


@dataclass()
class EventDataFrames:
    """Class for keeping different types of dataframes."""

    tracks: pd.DataFrame
    identities: pd.DataFrame
    pages: pd.DataFrame
    screens: pd.DataFrame
    groups: pd.DataFrame
    aliases: pd.DataFrame

    def __post_init__(self):
        for df in [
            self.tracks,
            self.identities,
            self.pages,
            self.screens,
            self.groups,
            self.aliases,
        ]:
            if not dataframe_util.empty(df):
                columns = df.columns.values
                for timestamp_field in event_fields.TIMESTAMP_FIELDS:
                    if timestamp_field in columns:
                        df[timestamp_field] = pd.to_datetime(df[timestamp_field])

        self.tracks["original_event"] = self.tracks["event"]
        self.tracks["event"] = self.tracks["event"].apply(
            lambda x: humps.decamelize(x.replace(" ", "")).lower()
        )

    def summary(self):
        return f"""
        tracks = {dataframe_util.row_count(self.tracks)}, 
        identities = {dataframe_util.row_count(self.identities)}, 
        pages = {dataframe_util.row_count(self.pages)}, 
        screens = {dataframe_util.row_count(self.screens)}, 
        groups = {dataframe_util.row_count(self.groups)}, 
        aliases = {dataframe_util.row_count(self.aliases)}"""

    def set_extra_timestamps(self, extra_timestamps: dict):
        for df in [
            self.tracks,
            self.identities,
            self.pages,
            self.screens,
            self.groups,
            self.aliases,
        ]:
            if not dataframe_util.empty(df):
                columns = df.columns.values
                for ts_name, tz in extra_timestamps.items():
                    if ts_name in columns:
                        raise Exception(f"Column with {ts_name} already exist")
                    logging.info(f"Creating new timestamp {ts_name} for zone {tz}")
                    df[ts_name] = df["timestamp"].dt.tz_convert(tz).dt.tz_localize(None)
                    # logging.info(f"{df['timestamp']}")
                    # logging.info(f"{df[ts_name]}")
                    # raise Exception("99")


class SendToWarehouseJob:
    """ Handles whole process to send files to warehouse """

    app_conf: AppConf
    source_dir: str
    warehouse_namespace: str
    warehouse_schema: str
    warehouses: List[wh.Warehouse]
    non_null_columns: List[str]

    def __init__(self, app_conf: AppConf, source_dir: str, warehouse_namespace: str):
        self.app_conf = app_conf
        self.source_dir = source_dir
        self.warehouse_namespace = warehouse_namespace
        self.warehouse_schema = humps.decamelize(self.warehouse_namespace)
        self.warehouses = []
        for warehouse_conf in app_conf.warehouses:
            self.warehouses.append(whf.get_warehouse(warehouse_conf))
        self.non_null_columns = [event_fields.RECEIVED_AT, event_fields.TIMESTAMP, event_fields.MESSAGE_ID] + list(
            self.app_conf.extra_timestamps.keys())

    def execute(self):
        file_names = [
            f for f in listdir(self.source_dir) if isfile(join(self.source_dir, f))
        ]
        file_paths = [self.source_dir + "/" + x for x in file_names]

        logging.info(f"Files to be sent to warehouses are : {file_paths}")

        self.process(file_paths)

    def process(self, file_paths):
        for file_path in file_paths:
            logging.info(f"Started processing {file_path}")
            file_df = self.get_file_df(file_path)

            logging.info(f"Removing columns = {self.app_conf.skip_fields}")
            file_df = file_df.drop(columns=self.app_conf.skip_fields)

            event_data_frames = self.break_down_by_type(file_df)

            event_data_frames.set_extra_timestamps(self.app_conf.extra_timestamps)
            self.store(event_data_frames)
            logging.info(f"Completed processing {file_path}")
        self.clean_up()

    def store(self, event_data_frames: EventDataFrames):
        self.store_identities(event_data_frames.identities)
        self.store_tracks(event_data_frames.tracks)
        self.store_screens(event_data_frames.screens)
        self.store_pages(event_data_frames.pages)
        self.store_groups(event_data_frames.groups)
        self.store_aliases(event_data_frames.aliases)

    def clean_up(self):
        for warehouse in self.warehouses:
            warehouse.close()

    def store_identities(self, identities_df):
        if not dataframe_util.empty(identities_df):
            col_types = dataframe_util.get_datatypes(identities_df)
            logging.debug(f"Col, Types = {col_types}")

            identities_df = dataframe_util.mark_nan_to_none(identities_df, col_types)

            self.ensure_table_structure(
                self.warehouse_schema,
                default_table_structure.IDENTITIES_TABLE,
                default_table_structure.IDENTITIES,
                col_types,
            )
            for warehouse in self.warehouses:
                warehouse.insert_df(self.warehouse_schema, "identities", identities_df)

            self.store_users(identities_df)

    def store_users(self, identities_df):
        users_df = identities_df.copy()
        users_df['ver'] = users_df['timestamp'].astype(int)

        col_types = dataframe_util.get_datatypes(users_df)
        logging.debug(f"Col, Types = {col_types}")

        self.ensure_users_table_structure(
            self.warehouse_schema,
            default_table_structure.USERS,
            col_types,
        )
        for warehouse in self.warehouses:
            warehouse.insert_df(self.warehouse_schema, default_table_structure.USERS_TABLE, users_df)

    def ensure_users_table_structure(self, schema, default_structure, col_types):
        table = default_table_structure.USERS_TABLE
        users_non_null_columns = self.non_null_columns + ['ver', 'user_id']
        logging.debug(f"default_structure = {default_structure}")
        for warehouse in self.warehouses:
            warehouse.create_schema(schema)
            warehouse.create_users_table(schema, default_structure, users_non_null_columns)

            table_col_types = warehouse.describe_table(schema, table)
            for col_name, col_type in col_types.items():
                if col_name not in table_col_types:
                    warehouse.add_column(schema, table, col_name, col_type, users_non_null_columns)

    def store_tracks(self, tracks_df):
        if not dataframe_util.empty(tracks_df):
            selected_col_df = self.select_columns(
                tracks_df,
                list(default_table_structure.TRACKS.keys()) + list(self.app_conf.extra_timestamps.keys()),
                default_table_structure.TRACKS_ALLOWED_FIELD_PREFIXES,
            )
            col_types = dataframe_util.get_datatypes(selected_col_df)
            logging.debug(f"Col, Types = {col_types}")

            selected_col_df = dataframe_util.mark_nan_to_none(selected_col_df, col_types)

            self.ensure_table_structure(
                self.warehouse_schema,
                default_table_structure.TRACKS_TABLE,
                default_table_structure.TRACKS,
                col_types,
            )
            for warehouse in self.warehouses:
                warehouse.insert_df(self.warehouse_schema, "tracks", selected_col_df)

            self.store_individual_events(tracks_df)

    def store_individual_events(self, tracks_df):
        all_events = sorted(tracks_df["event"].unique())
        for event in all_events:
            event_df = tracks_df[tracks_df["event"] == event].copy()
            event_col_types = dataframe_util.get_datatypes(event_df)
            logging.debug(f"Event = {event}, Col, Types = {event_col_types}")
            table = event
            if table in default_table_structure.DEFAULT_TABLES:
                table = f"esc_{table}"

            event_df = dataframe_util.mark_nan_to_none(event_df, event_col_types)
            self.ensure_table_structure(
                self.warehouse_schema,
                table,
                default_table_structure.TRACKS,
                event_col_types,
            )
            for warehouse in self.warehouses:
                warehouse.insert_df(self.warehouse_schema, table, event_df)

    def store_screens(self, screens_df):
        if not dataframe_util.empty(screens_df):
            col_types = dataframe_util.get_datatypes(screens_df)
            logging.info(f"Col, Types = {col_types}")

            screens_df = dataframe_util.mark_nan_to_none(screens_df, col_types)

            self.ensure_table_structure(
                self.warehouse_schema,
                default_table_structure.SCREENS_TABLE,
                default_table_structure.SCREENS,
                col_types,
            )
            for warehouse in self.warehouses:
                warehouse.insert_df(self.warehouse_schema, "identities", screens_df)

    def store_pages(self, pages_df):
        if not dataframe_util.empty(pages_df):
            col_types = dataframe_util.get_datatypes(pages_df)
            logging.info(f"Col, Types = {col_types}")

            pages_df = dataframe_util.mark_nan_to_none(pages_df, col_types)

            self.ensure_table_structure(
                self.warehouse_schema,
                default_table_structure.PAGES_TABLE,
                default_table_structure.PAGES,
                col_types,
            )
            for warehouse in self.warehouses:
                warehouse.insert_df(self.warehouse_schema, "identities", pages_df)

    def store_groups(self, groups_df):
        if not dataframe_util.empty(groups_df):
            col_types = dataframe_util.get_datatypes(groups_df)
            logging.debug(f"Col, Types = {col_types}")

            groups_df = dataframe_util.mark_nan_to_none(groups_df, col_types)

            self.ensure_table_structure(
                self.warehouse_schema,
                default_table_structure.GROUPS_TABLE,
                default_table_structure.GROUPS,
                col_types,
            )
            for warehouse in self.warehouses:
                warehouse.insert_df(self.warehouse_schema, "identities", groups_df)

    def store_aliases(self, aliases_df):
        if not dataframe_util.empty(aliases_df):
            col_types = dataframe_util.get_datatypes(aliases_df)
            logging.debug(f"Col, Types = {col_types}")

            aliases_df = dataframe_util.mark_nan_to_none(aliases_df, col_types)

            self.ensure_table_structure(
                self.warehouse_schema,
                default_table_structure.ALIASES_TABLE,
                default_table_structure.ALIASES,
                col_types,
            )
            for warehouse in self.warehouses:
                warehouse.insert_df(self.warehouse_schema, "identities", aliases_df)

    def ensure_table_structure(self, schema, table, default_structure, col_types):
        logging.debug(f"default_structure = {default_structure}")
        for warehouse in self.warehouses:
            warehouse.create_schema(schema)
            warehouse.create_table(schema, table, default_structure, self.non_null_columns)

            table_col_types = warehouse.describe_table(schema, table)
            for col_name, col_type in col_types.items():
                if col_name not in table_col_types:
                    warehouse.add_column(schema, table, col_name, col_type, self.non_null_columns)

    @staticmethod
    def select_columns(df, keep_columns, keep_columns_with_prefixes):
        col_names = df.columns.values
        selected_col_names = set()
        for col_name in col_names:
            if col_name in keep_columns:
                selected_col_names.add(col_name)

            if col_name.startswith(keep_columns_with_prefixes):
                selected_col_names.add(col_name)
        logging.debug(f"selected_col_names = {selected_col_names}")
        return df[selected_col_names]

    @staticmethod
    def get_file_df(file_path):
        data = []

        if file_path.endswith(".gz"):
            opener = gzip.open
        else:
            opener = open

        with opener(file_path, "r") as f:
            for line in f:
                event_json = json.loads(line)
                snake_cased_event_json = humps.decamelize(event_json)
                data.append(snake_cased_event_json)

        logging.debug(
            f"first 5 event json objects = {json.dumps(data[0:5], indent=4, default=str)}"
        )

        flattened_data = []
        for d in data:
            flattened_data.append(json_util.flatten_json(d))

        logging.debug(
            f"first 5 flattened event json objects = {json.dumps(flattened_data[0:5], indent=4, default=str)}"
        )

        df = pd.DataFrame(flattened_data)
        return df

    @staticmethod
    def break_down_by_type(df):
        event_data_frames = EventDataFrames(
            tracks=df[df["type"] == "track"].copy(),
            identities=df[df["type"] == "identify"].copy(),
            pages=df[df["type"] == "page"].copy(),
            screens=df[df["type"] == "screen"].copy(),
            groups=df[df["type"] == "group"].copy(),
            aliases=df[df["type"] == "alias"].copy(),
        )
        logging.info(f"Event Data Frames Summary = {event_data_frames.summary()}")
        return event_data_frames
